"""Client for the deepstream.io realtime web server"""
from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.client import connect


__all__ = ["connect"]


version = "0.1.1"
version_info = (0, 1, 1)
