from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

CONNECTION = 'C'
AUTH = 'A'
ERROR = 'X'
EVENT = 'E'
RECORD = 'R'
RPC = 'P'
PRESENCE = 'U'
PRIVATE = 'PRIVATE/'
